#%%
import matplotlib.pyplot as plt
from astropy.io import fits
import numpy as np
import math
import glob
from scipy.interpolate import interp1d
from statistics import stdev
from scipy.optimize import curve_fit

# run_name = input('Run name (all_frames or jvc):' )
run_name = 'newdepths'

def gauss(x, rv, sd, height, cont):
    y = cont+(height*np.exp(-(x-rv)**2/(2*sd**2)))
    return y
 
def findfiles(directory):

    filelist= glob.glob('%s*_%s.fits'%(directory, run_name))

    return filelist                       

def remove_reflex(velocities, spectrum, errors, phi, K, e, omega, v0):
    velo = v0 + K*(e*np.cos(omega)+np.cos(2*np.pi*phi+omega))
    adjusted_velocities = velocities-velo
    f2 = interp1d(adjusted_velocities, spectrum, kind='linear', bounds_error=False, fill_value='extrapolate')
    velocity_grid = np.arange(-20,20,0.82)
    adjusted_spectrum = f2(velocity_grid)
    f2 = interp1d(adjusted_velocities, errors, kind='linear', bounds_error=False, fill_value='extrapolate')
    adjusted_errors = f2(velocity_grid)
    
    return velocity_grid, adjusted_spectrum, adjusted_errors

def combineprofiles(spectra, errors, master, velocities):
    spectra = np.array(spectra)
    idx = np.isnan(spectra)
    shape_og = spectra.shape
    if len(spectra[idx])>0:
        spectra = spectra.reshape((len(spectra)*len(spectra[0]), ))
        for n in range(len(spectra)):
            if spectra[n] == np.nan:
                spectra[n] = (spectra[n+1]+spectra[n-1])/2
                if spectra[n] == np.nan:
                    spectra[n] = 0.
    spectra = spectra.reshape(shape_og)
    errors = np.array(errors)

    if master == 'no':

        weights_csv = np.genfromtxt('%sorder_weights.csv'%file_path, delimiter=',')
        orders = np.array(weights_csv[:len(spectra),0], dtype = int)
        weights_temp = np.array(weights_csv[:len(spectra):,1])
        spectra_to_combine = []
        errorss = []
        weights = []
        for i in orders:
            if np.sum(spectra[i-1])!=0:
                spectra_to_combine.append(list(spectra[i-1]))
                errorss.append(list(errors[i-1]))
                weights.append(weights_temp[i-1])
        weights = np.array(weights/sum(weights))

    else:
        spectra_to_combine = []
        weights=[]
        for n in range(0, len(spectra)):
            if np.sum(spectra[n])!=0:
                spectra_to_combine.append(list(spectra[n]))
                temp_err = np.array(errors[n, :])
                weight = (1/temp_err**2)
                weights.append(np.mean(weight))
        weights = np.array(weights/sum(weights))

    spectra_to_combine = np.array(spectra_to_combine)

    length, width = np.shape(spectra_to_combine)
    spectrum = np.zeros((1,width))
    spec_errors = np.zeros((1,width))

    for n in range(0,width):
        temp_spec = spectra_to_combine[:, n]
        spectrum[0,n]=sum(weights*temp_spec)/sum(weights)
        spec_errors[0,n]=(stdev(temp_spec)**2)*np.sqrt(sum(weights**2))

    spectrum = list(np.reshape(spectrum, (width,)))
    spec_errors = list(np.reshape(spec_errors, (width,)))

    return  spectrum, spec_errors, weights

def continuumfit(fluxes, wavelengths, errors, poly_ord):
        idx = wavelengths.argsort()
        wavelength = wavelengths[idx]
        fluxe = fluxes[idx]
        clipped_flux = []
        clipped_waves = []
        binsize =100
        for i in range(0, len(wavelength), binsize):
            waves = wavelength[i:i+binsize]
            flux = fluxe[i:i+binsize]
            indicies = flux.argsort()
            flux = flux[indicies]
            waves = waves[indicies]
            clipped_flux.append(flux[len(flux)-1])
            clipped_waves.append(waves[len(waves)-1])
        coeffs=np.polyfit(clipped_waves, clipped_flux, poly_ord)
        poly = np.poly1d(coeffs)
        fit = poly(wavelengths)
        flux_obs = fluxes/fit
        new_errors = errors/fit

        return flux_obs, new_errors


def classify(ccf, P, t, T):
    #working out if in- or out- of transit
    #print(t, b)
    deltaphi = t/(2*P)
    #print(deltaphi)
    phi = (((ccf[0].header['ESO DRS BJD'])-T)/P)%1
    #z = np.sqrt((a_rstar*(np.sin(2*math.pi*phi))**2)+b**2*(np.cos(2*math.pi*phi))**2)
    #part1=a_rstar*np.sin(2*math.pi*phi)**2
    #part2=b**2*np.cos(2*math.pi*phi)**2
    #z=np.sqrt(part1+part2)
    #print(z, phi)
    '''
    if z < 1+rp_rstar:
        result = 'in'
    else:result = 'out'
    '''
    phase = phi
    if phi < deltaphi:
        result ='in'
    elif phi > 1-deltaphi:
        result = 'in'
    else:result = 'out'
    #print(result)
    if phi>0.5:
        phi = phi-1

    return result, phi, phase
####################################################################################################################################################################

# Parameters for HD189733b

P=2.21857567 #Cegla et al, 2006 - days
i = 85.71*np.pi/180 #got quickly of exoplanet.eu
T=2454279.436714 #cegla et al,2006
t=0.076125 #Torres et al, 2008
e =0
omega=(np.pi/2)
K=0.20056 #km/s Boisse et al, 2009
v0=-2.2765#-0.1875 #km/s Boisse et al, 2009

a_rstar= 8.863 #Agol et al, 2010
rp_rstar= 0.15667 #Agol et al, 2010
u1=0.816
u2=0            #Sing et al, 2011
P=2.21857567 #Cegla et al, 2016 - days
T=2454279.436714 #cegla et al,2016
a_Rs = 8.786 #Cristo et al - 8.786
b=0.687 #Cristo et al, 2022
RpRs = 0.15667

# path = '/home/lsd/Documents/Starbase/novaprime/Documents/LSD_Figures/'
# save_path = '/home/lsd/Documents/Starbase/novaprime/Documents/LSD_Figures/'

path = '/Users/lucydolan/Starbase/newdepths/'
file_path = '/Users/lucydolan/Starbase/'

months = ['August2007',
          'July2007',
          'July2006',
          'Sep2006'
          ]

for month in months:
    filelist = findfiles('%s%s'%(path,month))
    velocities=np.arange(-21, 18, 0.82)

    frame_profiles = np.zeros((len(filelist), 3, len(np.arange(-20,20,0.82))))
    for frame_no in range(len(filelist)):
        file = fits.open(filelist[frame_no])
        order_errors = []
        order_profiles = []

        for order1 in range(0,71):

            profile_errors = file[order1].data[1]
            profile = file[order1].data[0]

            if order1>38 and order1<43:
                print('ignoring order %s'%order1)
                profile_errors = np.zeros(profile_errors.shape)
                profile = np.zeros(profile.shape)
            
            if order1==0:
                phase = file[order1].header['PHASE']
            elif file[order1].header['PHASE']!=file[order1-1].header['PHASE']:
                raise ValueError('Phase does not match for the same frame!')

            order_errors.append(profile_errors)
            order_profiles.append(profile)

        spectrum, errors, weights = combineprofiles(order_profiles, order_errors, 'no', velocities)
        
        final_velocities, spectrum, errors = remove_reflex(velocities, spectrum, errors, phase, K, e, omega, v0)

        frame_profiles[frame_no] = [[phase-np.round(phase)]*len(spectrum), spectrum, errors]

    phi = frame_profiles[:, 0, 0]
    z = np.sqrt( ( (a_Rs)*np.sin(2 * np.pi * phi) )**2 + ( b*np.cos(2. * np.pi * phi))**2)

    out_idx = (z>1+RpRs)
    if len(frame_profiles[out_idx, 0])>1:
        master_out_spec, master_out_errors, master_weights = combineprofiles(frame_profiles[out_idx, 1], frame_profiles[out_idx, 2], 'yes', final_velocities)
    else:
        master_out_spec = frame_profiles[:, 1]
        master_out_errors = frame_profiles[:, 2]

    master_out = np.array([master_out_spec, master_out_errors])

    #write in data
    hdu=fits.HDUList()
    
    #write in header
    for p in range(len(frame_profiles)):
        hdu.append(fits.PrimaryHDU(data=frame_profiles[p, 1:]))
        phase = frame_profiles[p, 0, 0]
        phi = phase
        z = np.sqrt( ( (a_Rs)*np.sin(2 * np.pi * phi) )**2 + ( b*np.cos(2. * np.pi * phi))**2)
        if z>1+RpRs:
            result = 'out'
        else: result = 'in'
        hdr=fits.Header()
        hdr['CRVAL1']=np.min(final_velocities)
        hdr['CDELT1']=final_velocities[1]-final_velocities[0]
        hdr['OBJECT']='HD189733b'
        hdr['NIGHT']='%s'%month
        hdr['K']=K
        hdr['V0']=v0
        hdr['PHASE']=phase
        hdr['RESULT']=result
        hdu[p].header=hdr
    
    hdu.append(fits.PrimaryHDU(data=master_out))
    hdr=fits.Header()
    hdr['CRVAL1']=np.min(final_velocities)
    hdr['CDELT1']=final_velocities[1]-final_velocities[0]
    hdr['OBJECT']='HD189733b'
    hdr['NIGHT']='%s'%month
    hdr['K']=K
    hdr['V0']=v0
    hdr['PHASE']='master out'
    hdr['RESULT']='out'
    hdu[p+1].header=hdr

    hdu.writeto('%s%s_master_out_LSD_profile.fits'%(path, month), output_verify='fix', overwrite = 'True')

    rvs = []
    for y in range(len(frame_profiles[:])):
        ACID_profile = frame_profiles[y, 1]
        ACID_errors = frame_profiles[y, 2]
        popt, pcov = curve_fit(gauss, final_velocities[abs(final_velocities)<5], ACID_profile[abs(final_velocities)<5], sigma = ACID_errors[abs(final_velocities)<5], absolute_sigma=True)
        perr = np.sqrt(np.diag(pcov))
        rvs.append([popt[0], perr[0]])

    rvs = np.array(rvs)
    plt.figure('RVs')
    phi = frame_profiles[:, 0, 0]
    z = np.sqrt( ( (a_Rs)*np.sin(2 * np.pi * phi) )**2 + ( b*np.cos(2. * np.pi * phi))**2)
    out_idx = (z>1+RpRs)
    plt.errorbar(frame_profiles[:, 0, 0], rvs[:, 0], rvs[:, 1], marker ='o', color = 'orange', linestyle='')
    plt.show()



#%%
from astropy.io import fits
import numpy as np
import matplotlib.pyplot as plt
import glob, os, sys
from pathlib import Path
from time import time
from beartype.roar import BeartypeCallHintParamViolation
SCRIPT_DIR = Path(__file__).resolve().parent
PROJECT_ROOT = next(p for p in SCRIPT_DIR.parents if (p / "pyproject.toml").exists())
sys.path.append(str(PROJECT_ROOT))
os.chdir(PROJECT_ROOT)
from src import  ACID_code as acid
np.random.seed(0) # Set random seed for reproducibility in tests
start = time()
skips = 5

def legacy_test():

    def quickstart():
        spec_file = fits.open('example/sample_spec_1.fits')

        wavelength = spec_file[0].data[::skips]   # Wavelengths in Angstroms
        spectrum = spec_file[1].data[::skips]     # Spectral Flux
        error = spec_file[2].data[::skips]        # Spectral Flux Errors
        sn = spec_file[3].data           # SN of Spectrum

        linelist = 'example/example_linelist.txt' # Insert path to line list

        # choose a velocity grid for the final profile(s)
        deltav = acid.calc_deltav(wavelength)  # velocity pixel size must not be smaller than the spectral pixel size
        velocities = np.arange(-25, 25, deltav)

        # run ACID function
        result = acid.ACID(wavelength, spectrum, error, linelist, sn, velocities, nsteps=2000, parallel=False)

        # extract profile and errors
        profile = result[0, 0]
        profile_error = result[0, 1]

        # plot results
        plt.figure()
        plt.errorbar(velocities, profile, profile_error)
        plt.xlabel('Velocities (km/s)')
        plt.ylabel('Flux')
        plt.show()
        return result

    def multiple_frames():

        # finds sample files in 'example directory'. Each file is a different frame.
        files = glob.glob('example/sample_spec_*.fits')

        # create lists for wavelengths, spectra, errors and sn for all frames
        wavelengths = []
        spectra = []
        errors = []
        sns = []

        for file in files:
            spec_file = fits.open('%s'%file)

            wavelengths.append(spec_file[0].data[::skips])    # Wavelengths in Angstroms
            spectra.append(spec_file[1].data[::skips])        # Spectral Flux
            errors.append(spec_file[2].data[::skips])         # Spectral Flux Errors
            sns.append(float(spec_file[3].data[0]))     # SN of Spectrum

        linelist = 'example/example_linelist.txt' # Insert path to line list

        # choose a velocity grid for the final profile(s)
        deltav = acid.calc_deltav(wavelengths[0])
        velocities = np.arange(-25, 25, deltav)

        # run ACID function
        result = acid.ACID(wavelengths, spectra, errors, linelist, sns, velocities, nsteps=2000)

        # plot results
        plt.figure()

        for frame in range(len(files)):
            profile = result[frame, 0]
            profile_error = result[frame, 1]
            plt.errorbar(velocities, profile, profile_error, label = '%s'%frame)

        plt.xlabel('Velocities (km/s)')
        plt.ylabel('Flux')
        plt.legend()
        plt.show()
        return result

    _ = quickstart() # generic quick test
    _ = multiple_frames()

def class_test():
    def classes_test(skips, nsteps):
        spec_file = fits.open('example/sample_spec_1.fits')

        wavelength = spec_file[0].data   # Wavelengths in Angstroms
        spectrum = spec_file[1].data     # Spectral Flux
        error = spec_file[2].data        # Spectral Flux Errors
        sn = spec_file[3].data           # SN of Spectrum

        linelist = 'example/example_linelist.txt' # Insert path to line list

        # choose a velocity grid for the final profile(s)
        velocities = np.arange(-25, 25, 0.82)

        # run ACID function
        Acid = acid.Acid(velocities=velocities, linelist=linelist)
        result = Acid.ACID(wavelength, spectrum, error, sn, nsteps=nsteps, skips=skips)
        return result

    nsteps1 = 5000
    nsteps2 = 3000
    result = classes_test(skips=3, nsteps=nsteps1) # test the classes and result handling, use lower skips
    os.makedirs("tests/test_data", exist_ok=True) # create test data directory if it doesn't exist
    result.save("tests/test_data/classes_test_data.pkl", "tests/test_data/classes_test_sampler.h5")
    result = acid.Result.load("tests/test_data/classes_test_data.pkl")
    result.plot_corner()
    result.plot_profiles()
    result.plot_walkers()
    result.plot_forward_model()
    result.continue_sampling(nsteps=nsteps2)
    assert result.sampler.get_chain().shape[0] == nsteps1 + nsteps2 and result.data.nsteps == nsteps1 + nsteps2, \
    f"Continue sampling did not add the correct number of steps to the chain.\n" \
    f"Expected {nsteps1 + nsteps2}, got {result.sampler.get_chain().shape[0]} for sampler and {result.data.nsteps} for data."  
    result.plot_walkers()
    result.plot_autocorrelation()
    result.plot_acf()
    acid.Profiles(velocities=np.arange(-25, 25, 0.82), flux=result[0,0]).plot_fit("all")
    result.plot_profiles()
    del result # clean up memory after test

def verbosity_test():
    def no_verbosity():
        spec_file = fits.open('example/sample_spec_1.fits')

        wavelength = spec_file[0].data[::skips]   # Wavelengths in Angstroms
        spectrum = spec_file[1].data[::skips]     # Spectral Flux
        error = spec_file[2].data[::skips]        # Spectral Flux Errors
        sn = spec_file[3].data           # SN of Spectrum

        linelist = 'example/example_linelist.txt' # Insert path to line list

        # choose a velocity grid for the final profile(s)
        deltav = acid.calc_deltav(wavelength)  # velocity pixel size must not be smaller than the spectral pixel size
        velocities = np.arange(-25, 25, deltav)

        # run ACID function
        Acid = acid.Acid(velocities=velocities, linelist=linelist, verbose=False)
        result = Acid.ACID(wavelength, spectrum, error, sn, nsteps=2000)
        return result

    def high_verbosity():
        spec_file = fits.open('example/sample_spec_1.fits')

        wavelength = spec_file[0].data[::skips]   # Wavelengths in Angstroms
        spectrum = spec_file[1].data[::skips]     # Spectral Flux
        error = spec_file[2].data[::skips]        # Spectral Flux Errors
        sn = spec_file[3].data           # SN of Spectrum

        linelist = 'example/example_linelist.txt' # Insert path to line list

        # choose a velocity grid for the final profile(s)
        deltav = acid.calc_deltav(wavelength)  # velocity pixel size must not be smaller than the spectral pixel size
        velocities = np.arange(-25, 25, deltav)

        # run ACID function
        Acid = acid.Acid(velocities=velocities, linelist=linelist, verbose=3)
        result = Acid.ACID(wavelength, spectrum, error, sn, nsteps=2000)
        return result
    
    print("Starting no verbosity, until no verbosity is printed, nothing should be output")
    res_nv = no_verbosity()
    res_nv.continue_sampling(nsteps=1000)
    del res_nv # clean up memory after test
    print("End no verbosity, starting high verbosity")
    res_hv = high_verbosity()
    del res_hv # clean up memory after test

def deterministic_test():
    def deterministic_profile_fit(skips=5, nsteps=2000):
        files = glob.glob('example/sample_spec_*.fits')

        # create lists for wavelengths, spectra, errors and sn for all frames
        wavelengths = []
        spectra = []
        errors = []
        sns = []

        for file in files:
            spec_file = fits.open('%s'%file)

            wavelengths.append(spec_file[0].data)    # Wavelengths in Angstroms
            spectra.append(spec_file[1].data)        # Spectral Flux
            errors.append(spec_file[2].data)         # Spectral Flux Errors
            sns.append(float(spec_file[3].data[0]))     # SN of Spectrum

        linelist = 'example/example_linelist.txt' # Insert path to line list

        # choose a velocity grid for the final profile(s)
        velocities = np.arange(-25, 25, 0.82)

        # run ACID function
        Acid = acid.Acid(velocities=velocities, linelist=linelist)
        result = Acid.ACID(wavelengths, spectra, errors, sns, nsteps=nsteps, parallel=True, deterministic_profile=False, skips=skips)
        result.plot_walkers()
        return result
    
    res_deterministic_fit = deterministic_profile_fit(skips=3, nsteps=2000)
    res = res_deterministic_fit
    res.plot_walkers()
    res.plot_profiles()
    res.plot_corner()
    res.plot_autocorrelation()
    res.plot_acf()

def data_and_convergence_test():
    spec_file = fits.open('example/sample_spec_1.fits')

    wavelength = spec_file[0].data   # Wavelengths in Angstroms
    spectrum = spec_file[1].data     # Spectral Flux
    error = spec_file[2].data        # Spectral Flux Errors
    sn = spec_file[3].data           # SN of Spectrum

    linelist = 'example/example_linelist.txt' # Insert path to line list

    # choose a velocity grid for the final profile(s)
    velocities = np.arange(-25, 25, 0.82)

    # run ACID function
    Acid = acid.Acid(velocities=velocities, linelist=linelist)
    result = Acid.ACID(wavelength, spectrum, error, sn, nsteps=2000, skips=skips)

    data = result.data

    Acid = acid.Acid(data=data) # data does not store the sampler
    result2 = Acid.ACID(nsteps=1000, parallel=True)

    assert result2.data.nsteps == 3000, "Continue sampling did not add the correct" \
    " number of steps to the chain when using data class."

    Acid.ACID(max_steps=5000) # test new convergence function

    # test convergence check interval
    Acid = acid.Acid(data=data)
    data.config.deterministic_profile = True
    Acid.ACID(max_steps=5000)

    return

def test_edge_cases():
    spec_file = fits.open('example/sample_spec_1.fits')

    wavelength = spec_file[0].data   # Wavelengths in Angstroms
    spectrum = spec_file[1].data     # Spectral Flux
    error = spec_file[2].data        # Spectral Flux Errors
    # sn = spec_file[3].data         # SN of Spectrum

    linelist = 'example/example_linelist.txt' # Insert path to line list
    full_linelist = np.genfromtxt('%s'%linelist, skip_header=4, delimiter=',', usecols=(1,9), invalid_raise=False)
    wl = full_linelist[:,0]
    depths = full_linelist[:,1]
    wl[199] = np.nan # introduce a nan value to test handling of nans in linelist
    depths[200] = np.nan # introduce a nan value to test handling of nans in linelist
    spectrum[201] = np.nan # introduce a nan value to test handling of nans in spectrum
    error[190] = np.nan # introduce a nan value to test handling of nans in error array

    # choose a velocity grid for the final profile(s)
    # velocities = np.arange(-25, 25, 0.82)

    # run ACID function
    Acid = acid.Acid(linelist=linelist)
    result = Acid.ACID(wavelength, spectrum, error, nsteps=2000, 
                       skips=skips, parallel=False, deterministic_profile=True)

    data = result.data
    # continue sampling
    Acid = acid.Acid(data=data)
    result = Acid.ACID(max_steps=5000) # test continue sampling with no parallelisation

    result.continue_sampling(max_steps=5000)
    result.plot_profiles()
    result.plot_walkers()
    result.plot_autocorrelation()

    # Plot continuum fit with the data class
    result.data.plot_continuum_fit("initial")
    result.data.plot_continuum_fit("masked")
    result.data.plot_residual_masking()

    # per pixel snr, and no run_mcmc test
    sn = np.random.normal(loc=100, scale=10, size=spectrum.shape) # create a random sn array with the same shape as the spectrum
    Acid = acid.Acid(linelist=linelist)
    result = Acid.ACID(wavelength, spectrum, error, sn, nsteps=2000, run_mcmc=False,
                       skips=skips, parallel=False, deterministic_profile=True)
    assert result is None, "When run_mcmc is set to False, the ACID function should return None, but it did not."

    # Guess sn for multiple frames, plot multiple frames
    files = glob.glob('example/sample_spec_*.fits')
    wavelengths = []
    spectra = []
    errors = []
    sns = []
    for file in files:
        spec_file = fits.open('%s'%file)
        wavelengths.append(spec_file[0].data[::skips])    # Wavelengths in Angstroms
        spectra.append(spec_file[1].data[::skips])        # Spectral Flux
        errors.append(spec_file[2].data[::skips])         # Spectral Flux Errors
        sns.append(float(spec_file[3].data[0]))     # SN of Spectrum
    deltav = acid.calc_deltav(wavelengths[0])
    velocities = np.arange(-25, 25, deltav)
    result = acid.ACID(wavelengths, spectra, errors, linelist, sns, velocities, nsteps=2000)
    result.plot_profiles()

    # Test using a different moveset:
    Acid = acid.Acid(data=data)
    data.config.moves = [("StretchMove", 0.6, {}), ("DEMove", 0.2)]
    result = Acid.ACID(max_steps=5000)
    result.plot_walkers()

    # Test plotting masking lines
    result.data.config.plot_masking_lines()

    # Test plotting linelist
    result.data.plot_linelist()

    # Own telluric lines
    pass

def test_data_and_datalist():
    # Test that the data class can be initialized with a datalist, 
    # and that the datalist is correctly stored in the data class
    files = glob.glob('example/sample_spec_*.fits')
    wavelengths = []
    spectra = []
    errors = []
    sn = []
    for file in files:
        spec_file = fits.open('%s'%file)
        wavelengths.append(spec_file[0].data)    # Wavelengths in Angstroms
        spectra.append(spec_file[1].data)        # Spectral Flux
        errors.append(spec_file[2].data)         # Spectral Flux Errors
        sn.append(spec_file[3].data)
    linelist = 'example/example_linelist.txt'
    velocities = np.arange(-25, 25, acid.calc_deltav(wavelengths[0]))
    datalist = acid.DataList(wavelengths, spectra, errors, sn, velocities, linelist, skips=skips)
    order_range = [20,21,22]
    configs = acid.Config(max_steps=5000)
    configs = [acid.Config(max_steps=5000) for _ in range(3)]
    configs[1].update_hipri(poly_ord=4)
    datalist = acid.DataList(
        wavelengths,
        spectra,
        errors,
        sn,
        velocities,
        linelist,
        verbose=2,
        save_dir="tests/test_data/datalist/",
        order_range=order_range,
        config=configs,
        cores=10,
        )
    datalist[22].config.update_hipri(poly_ord=4)
    datalist.run_ACID(allow_overwrite=True)
    datalist[22].result.plot_profiles()
    datalist.combine_profiles(exclude=21)
    datalist.plot_combined_profile()
    datalist.fit_profile()
    datalist.save()

def saves_and_loads():

    spec_file = fits.open('example/sample_spec_1.fits')
    wavelength = spec_file[0].data[::skips]   # Wavelengths in Angstroms
    spectrum = spec_file[1].data[::skips]     # Spectral Flux
    error = spec_file[2].data[::skips]        # Spectral Flux Errors
    sn = spec_file[3].data           # SN of Spectrum
    linelist = 'example/example_linelist.txt' # Insert path to line list
    deltav = acid.calc_deltav(wavelength)
    velocities = np.arange(-25, 25, deltav)
    result = acid.ACID(wavelength, spectrum, error, linelist, sn, velocities, nsteps=2000)
    result.save("tests/test_data/result_test.pkl")
    result.load("tests/test_data/result_test.pkl")
    result.plot_profiles()

    datalist = acid.DataList.load("tests/test_data/datalist/")
    datalist.excluded_orders = [21]
    datalist.plot_combined_profile()

    datalist = acid.DataList.load("tests/test_data/datalist/datalist.pkl")
    datalist.plot_combined_profile()
    datalist.fit_profile()

    import shutil
    shutil.move("tests/test_data/datalist/datalist.pkl", "tests/test_data/datalist.pkl")
    datalist = acid.DataList.load("tests/test_data/datalist/")
    datalist.plot_combined_profile()
    datalist.fit_profile()
    shutil.move("tests/test_data/datalist.pkl", "tests/test_data/datalist/datalist.pkl")

    # Testing regular result
    # Get back a result to test it with
    data = acid.Data()
    spec_file = fits.open('example/sample_spec_1.fits')
    wavelength = spec_file[0].data   # Wavelengths in Angstroms
    spectrum = spec_file[1].data     # Spectral Flux
    error = spec_file[2].data        # Spectral Flux Errors
    sn = spec_file[3].data           # SN of Spectrum
    linelist = 'example/example_linelist.txt' # Insert path to line list
    deltav = acid.calc_deltav(wavelength)
    velocities1 = np.arange(-25, 25, deltav)
    Acid = acid.Acid(velocities=velocities1, linelist=linelist, data=data)
    result = Acid.ACID(wavelength, spectrum, error, sn, nsteps=2000, skips=3)

    # Check result survives multiple saves and loads and can continue sampling after loading
    data_path = "tests/test_data/result_test.pkl"
    sampler_path = "tests/test_data/result_test_sampler.h5"
    result.save(data_path, sampler_path)
    result = acid.Result.load(data_path)
    result.plot_corner() # requires sampler
    result.continue_sampling(nsteps=1000)
    result.save(data_path)
    result = acid.Result.load(data_path)
    result.plot_corner() # requires sampler

def error_and_inputs_handlings():
    orderd = acid.Config.defaults["order"] # is 0
    order1 = 1
    order2 = 2
    try:
        cfg = acid.Config(orderd=orderd) # test typo in config kwargs
    except KeyError as e:
        print("Passed typo in config kwargs test:", e)
    config = acid.Config()
    assert config.order == orderd, "Config did not initialize order to default correctly."
    config.update_lowpri(order=order1)
    assert config.order == order1, "Config did not update order correctly using update_lowpri."
    config.update_hipri(order=order2)
    assert config.order == order2, "Config did not update order correctly using update_hipri."
    config.update_lowpri(order=order1)
    assert config.order == order2, "Config did not prioritise hipri updates over lowpri updates."

    # Same tests but now trying for verbose (which is property and acts slightly differently)
    verbose_d = acid.Config.defaults["verbose"] # is 1
    verbose1 = 0
    verbose2 = 3
    try:
        cfg = acid.Config(verbosee=verbose_d) # test typo in config kwargs
    except KeyError as e:
        print("Passed typo in config kwargs test for verbose:", e)
    config = acid.Config()
    assert config.verbose == verbose_d, "Config did not initialize verbose to default correctly."
    config.update_lowpri(verbose=verbose1)
    assert config.verbose == verbose1, "Config did not update verbose correctly using update_lowpri."
    config.update_hipri(verbose=verbose2)
    assert config.verbose == verbose2, "Config did not update verbose correctly using update_hipri."
    config.update_lowpri(verbose=verbose1)
    assert config.verbose == verbose2, "Config did not prioritise hipri updates over lowpri updates for verbose."
    acid.Config.print_defaults()

    # Lastly check config also overwrites properly
    cfg1 = acid.Config(order=order1)
    cfg2 = acid.Config(order=order2)
    data = acid.Data()
    data.config = cfg1
    assert data.config.order == order1, "Data did not update config order correctly when directly assigning a config object."
    data.config = cfg2
    assert data.config.order == order2, "Data did not update config order correctly when directly assigning a config object the second time, should overwrite previous config."

    # Test out the __setattr__
    cfg = acid.Config()
    cfg.order = 3
    assert cfg.order == 3, "Config did not set order to 3 correctly using __setattr__."
    cfg.order = None
    assert cfg.order == 3, "Config did not keep order on 3 correctly when given None using __setattr__."
    cfg = acid.Config()
    cfg.order = None
    assert cfg.order == 0, "Config did not initialize order to default correctly when using __setattr__."

    # Test it also works with with properties like verbose, and that it catches typos in attribute names
    cfg = acid.Config()
    cfg.verbose = 1
    assert cfg.verbose == 1, "Config did not set verbose to 1 correctly using __setattr__."
    try:
        cfg.verbosed = 2
    except AttributeError as e:
        print(f"Caught typo in config attribute name using __setattr__: {e}")
    

    
    # Test a data input to config
    try:
        cfg = acid.Config(linelist=[[1,2,3], [4,5,6]])
    except Exception as e:
        print(f"Caught the Data input into config: {e}")

    # Now test for Data instance which should always overwrite
    data = acid.Data()
    try:
        data.linelist = [[1,2,3], [4,5,6]]
    except ValueError as e:
        print(f"Caught attempt to set fully invalid linelist: {e}")
    data.linelist = [[1,2,3], [0.5,0.5,0.5]]
    assert np.array_equal(data.linelist["wavelengths"], [1,2,3])
    assert np.array_equal(data.linelist["depths"], [0.5,0.5,0.5])
    data.linelist = [[1,2,3], [0.6,0.6,0.6]]
    assert np.array_equal(data.linelist[0], [1,2,3])
    assert np.array_equal(data.linelist[1], [0.6,0.6,0.6])

    # Test verbosity type inputs
    config = acid.Config()
    config.verbose = 0
    assert config.verbose == 0, "Config did not set verbose to 0 correctly."
    config.verbose = 3
    assert config.verbose == 3, "Config did not set verbose to 3 correctly."
    try:
        config.verbose = -1
    except ValueError as e:
        print(f"Caught invalid verbose value: {e}")
    try:
        config.verbose = 4
    except ValueError as e:
        print(f"Caught invalid verbose value: {e}")
    config.verbose = True
    assert config.verbose == 2, "Config did not set verbose to 2 correctly when given True."
    config.verbose = False
    assert config.verbose == 0, "Config did not set verbose to 0 correctly when given False."
    config.verbose = "high"
    assert config.verbose == 3, "Config did not set verbose to 3 correctly when given 'high'."
    config.verbose = "low"
    assert config.verbose == 1, "Config did not set verbose to 1 correctly when given 'low'."
    config.verbose = "medium"
    assert config.verbose == 2, "Config did not set verbose to 2 correctly when given 'medium'."
    try:
        config.verbose = "invalid"
    except ValueError as e:
        print(f"Caught invalid verbose string: {e}")
    config.verbose = "off"
    assert config.verbose == 0, "Config did not set verbose to 0 correctly when given 'off'."
    config.verbose = None
    assert config.verbose == 0, "Config did not keep verbose on 0 correctly when given None."

    # Now test the linelist input types:
    data = acid.Data()
    f = 0.5
    try:
        data.linelist = [[1,2,3], [f,f]] # test different length inputs
    except ValueError as e:
        print(f"Caught different length linelist inputs: {e}")
    try:
        data.linelist = [["1","2","3"], ["hi","0.5","0.5"]]
    except BeartypeCallHintParamViolation as e:
        print(f"Caught non-numeric linelist depths with beartype: {e}")
    data.linelist = {"wavelengths": [1,2,3], "depths": [f,f,f]} # test dict input
    assert np.array_equal(data.linelist["wavelengths"], [1,2,3])
    assert np.array_equal(data.linelist["depths"], [f,f,f])
    data.linelist = {"wavelengths": [1,2,3], "depths": [f,f,f], "extra_key": "ignored"} # test dict input with extra keys
    assert np.array_equal(data.linelist["wavelengths"], [1,2,3])
    assert np.array_equal(data.linelist["depths"], [f,f,f])
    data.linelist = None # should not update since None is not a valid linelist
    assert np.array_equal(data.linelist["wavelengths"], [1,2,3])
    assert np.array_equal(data.linelist["depths"], [f,f,f])

    # Check velocities does overwrite
    data.velocities = np.arange(-25, 25, 0.82)
    data.velocities = np.arange(-30, 30)
    assert np.array_equal(data.velocities, np.arange(-30, 30)), "Data instance did not update velocities correctly."
    data.velocities = np.arange(-30, 30) # should do nothing

    # Check linelist survives the save/load
    data = acid.Data()
    data.linelist = [[1,2,3], [0.5,0.5,0.5]]
    data.save("tests/test_data/data_test.pkl")
    data_loaded = acid.Data.load("tests/test_data/data_test.pkl")
    assert np.array_equal(data_loaded.linelist["wavelengths"], [1,2,3])
    assert np.array_equal(data_loaded.linelist["depths"], [0.5,0.5,0.5])

    # Check masking lines
    masks = data.config.masking_lines.get_masks(x=np.array([1,2,3]))
    assert len(masks) == 3, "Masking lines did not return the correct number of masks."
    assert type(masks) == list, "Masking lines did not return masks as numpy arrays."
    masks = data.config.masking_lines.get_masks(x=np.array([1,2,3]), with_names=True)
    assert type(masks) == dict, "Masking lines did not return masks as a dict when with_names is True."

    # Check data reset
    spec_file = fits.open('example/sample_spec_1.fits')
    wavelength = spec_file[0].data   # Wavelengths in Angstroms
    spectrum = spec_file[1].data     # Spectral Flux
    error = spec_file[2].data        # Spectral Flux Errors
    sn = spec_file[3].data           # SN of Spectrum
    data = acid.Data()
    data.set_inputs(wavelength, spectrum, error, sn)
    wl1 = data.wavelengths["input"]
    sp1 = data.flux["input"]
    assert np.array_equal(wl1[0], wavelength) and np.array_equal(sp1[0], spectrum), "Data instance did not set inputs correctly."
    try:
        data.set_inputs([24], [0.5], [0.1], [10])
    except ValueError as e:
        print(f"Caught attempt to set len(wavelengths) <= 1: {e}")
    data.alpha = 0.5
    data.set_inputs([24, 25], [0.5, 0.6], [0.1, 0.1], [10, 20])
    assert data.sn["input"].ndim + 1 == data.wavelengths["input"].ndim, "Data instance did not set sn with correct dimensions when given multiple frames."
    assert data.alpha is None, "Data instance did not reset alpha to None when setting new inputs with multiple frames."
    data.set_inputs(wavelength, spectrum, input_sn=sn) # reset to original inputs for any future tests
    assert data.errors["input"].ndim == data.wavelengths["input"].ndim, "Data instance did not calculate errors from the SN"

def LSD():
    spec_file = fits.open('example/sample_spec_1.fits')
    wavelength = spec_file[0].data[::5]   # Wavelengths in Angstroms
    spectrum = spec_file[1].data[::5]     # Spectral Flux
    error = spec_file[2].data[::5]        # Spectral Flux Errors
    sn = spec_file[3].data                # SN of Spectrum
    linelist = 'example/example_linelist.txt' # Insert path to line list
    deltav = acid.calc_deltav(wavelength)
    velocities1 = np.arange(-25, 25, deltav)
    Acid = acid.Acid(velocities=velocities1, linelist=linelist)
    data = Acid.data

    # Try LSD in OD and flux
    lsd = acid.LSD(data=data, OD=True)
    lsd.run_LSD(wavelength, spectrum, error, sn)

    # And in flux
    data.reset()
    lsd = acid.LSD(data=data, OD=False)
    lsd.run_LSD(wavelength, spectrum, error, sn)

print("Starting tests, this will take a 2-5 minutes to run, and a bunch of output will be printed.")

# The first five tests use legacy ACID inputs and calls
legacy_test()

# Now test classes
class_test()

# Test verbosities
verbosity_test()

# Test deterministic profile fit
deterministic_test()

# Add a skipping calculations using the data class test
data_and_convergence_test()

# Test edge cases, including no parallelization
test_edge_cases()

# Test datalist and data class integration
test_data_and_datalist()

# Test saves and loads
saves_and_loads()

# Test error handlings
error_and_inputs_handlings()

# Test LSD
LSD()

print("All tests passed!")
print(f"Total time: {time() - start:.2f} seconds")

# %%



"""
Custom error classes for ACID. To be used mainly for handling specific errors in the DataList run_ACID method.
"""

class LineListRangeError(ValueError):
    """Custom error for when no lines in the linelist are within the wavelength range of the observed spectrum."""
    pass

class ContinuumError(ValueError):
    """Custom error for when the continuum fit results in negative fluxes or errors."""
    pass

class SNCutError(ValueError):
    """Custom error for when the S/N cut results in no valid pixels."""
    pass


"""
This script contains functions to load data from different fits files and instruments to be used with Acid.
Each function will load a data object that can be directly input into the ACID initialisation.
"""
from __future__ import annotations
from beartype import beartype
import numpy as np
from .data import Data, DataList
from . import utils
from .utils import Array2D
from astropy.io import fits

@beartype
class Load:

    def __init__(
            self,
            instrument    : str,
            file          : str,
            blaze_file    : str|None     = None,
            blaze_profile : Array2D|None = None
            ):
        """
        Parameters
        ----------
        instrument : str
            The name of the instrument the data was taken with. This will determine how the data is loaded and configured.
            Current options are "HARPS", "HARPS-N", "GEMINI-GHOST", "GEMINI-GRACES"
        file : str
            The path to the fits file containing the data to be loaded.
        blaze_file : str, optional
            The path to the fits file containing the blaze profile for the data to be loaded.
            If not provided, no blaze correction is applied. By default, None
        blaze_profile : Array2D, optional
            Instead load the blaze profile yourself and input it here. This will override the blaze_file input if 
            both are provided. By default, None
        """
        raise NotImplementedError("This class is still in development and not yet ready for use. Please check back later.")

        self.file = file
        self.blaze_file = blaze_file

        instruments = {
            "harps": {
                "function": self.HARPS,
                "accepted_names": ["harps", "harps-s", "harps-south"]
            },
            "harps-n": {
                "function": self.HARPS_N,
                "accepted_names": ["harps-n", "harps-north", "harpsn", "harps_north", "harps_n"]
            },
            "gemini_ghost": {
                "function": self.GEMINI_GHOST,
                "accepted_names": ["gemini_ghost", "ghost", "gemini-ghost"]
            },
            "gemini_graces": {
                "function": self.GEMINI_GRACES,
                "accepted_names": ["gemini_graces", "graces", "gemini-graces"]
            },
        }

        for k, v in instruments.items():
            if instrument.lower() in v["accepted_names"]:
                self.loader = v["function"]
                self.instrument = k
                break
        if not hasattr(self, "loader"):
            raise ValueError(f"Instrument '{instrument}' not recognised. Accepted instruments are: {', '.join(instruments.keys())}.")
        self.instrument = instrument # instrument is now an accepted name

        # Save exceptions now
        self.load_exception = f"Error loading data for the '{self.instrument}' instrument. \n" \
        f"Please check the file format and ensure it follows standard routines for extracting with fits formats. \n" \
        f"If you believe this is a standard format and are still getting this error, \n " \
        f"please open an issue on the ACID GitHub repository with details of the file format and the error message you received." \
        f"Exception: \n"

        self.data = Data()
        # Load data into class
        self.hdul = fits.open(file) # standard astropy error catching will catch if this fails
        self.loader(self.hdul)

        # If blaze profile is provided, load it
        if blaze_profile is not None:
            self.data.blaze = blaze_profile
        elif blaze_file is not None:
            pass # TODO: load blaze file and save to self.data.blaze or tailor blaze files to each instrument

    def to_datalist(self, **kwargs) -> DataList:
        datalist = DataList.create(
            wavelengths = self.wavelengths,
            flux        = self.flux,
            errors      = self.errors,
            sn          = self.sn,
            order_range = self.order_range,
            **kwargs,
        )
        return datalist

    def HARPS(self):
        # La Silla HARPS spectra
        try:
            hdr = self.hdul[0].header
            data = self.hdul[0].data
            wavelengths = self.hdul[1].data
        except Exception as e:
            raise ValueError(self.load_exception + str(e))


        if self.blaze_profile is not None:
            try:
                blaze_hdu = fits.open(self.blaze_file)
                blaze_profiles = np.array(blaze_hdu[0].data)
                self.flux = data / blaze_profiles
            except:
                print("Error loading blaze profile. No blaze correction applied.")

        return

    def HARPS_N(self):
        hdul = fits.open(self.file)
        hdr = hdul[0].header
        flux = hdul[1].data
        errors = hdul[2].data
        dq = hdul[3].data
        wavelengths = hdul[5].data
        sn = np.array([hdr[f"HIERARCH TNG QC ORDER{i+1} SNR"] for i in range(len(flux))])

        pass

    def GEMINI_GHOST(self):
        # Gemini GHOST spectra
        pass

    def GEMINI_GRACES(self):
        # Gemini GRACES spectra
        pass

    def CARMENES(self):
        # Calar Alto CARMENES spectra
        pass

    def VLT_UVES(self):
        # La Silla UVES spectra
        pass

    def HARPS3(self):
        # La Palma HARPS3 spectra
        pass

    def ESPRESSO(self):
        # La Silla ESPRESSO spectra
        pass




from __future__ import annotations
from time import time
import numpy as np
import matplotlib.pyplot as plt
import corner, sys, os, pickle, warnings, contextlib, functools, inspect, psutil
from emcee import EnsembleSampler
import emcee.backends.backend as emceebackend
from beartype import beartype
from scipy.interpolate import interp1d
from numpy.polynomial import polynomial as P
from .lsd import LSD
from . import mcmc
from . import utils
from .data import Data
from .utils import IntLike, Scalar
#TODO: utils.set_dict_defaults for plots

warnings.filterwarnings("ignore")

def _require_profiles(method):
    # Make sure all results are processed before calling method
    @functools.wraps(method)
    def wrapper(self, *args, **kwargs):
        if not self.data.complete: # complete is flag for if profiles have been made
            name = method.__qualname__
            if self.sampler is not None:
                if self.config.verbose>0:
                    print(f"Note: The Result object was created without the profiles processed. " \
                        f"Running {name} requires all results to be processed, " \
                        "so process_results() will now be called...")
                self.process_results()
            else:
                error = f"Cannot call {name}. The profiles attribute is not available, and no " \
                "sampler object is available to process results. Please pass an Acid/Data " \
                "instance after running ACID to the results init."
                raise ValueError(error)
        return method(self, *args, **kwargs)
    return wrapper

def _require_sampler(method):
    # Make sure sampler object is available before calling method
    sig = inspect.signature(method)
    @functools.wraps(method)
    def wrapper(self, *args, **kwargs):
        bound = sig.bind_partial(self, *args, **kwargs)
        inputted_sampler = bound.arguments.get("sampler", None)
        self.initiate_sampler(inputted_sampler, _method_name=method.__qualname__)
        return method(self, *args, **kwargs)
    return wrapper

@beartype
class Result:
    """
    Class to handle the results from the Acid MCMC sampling, and results processing. Fundamentally, this
    class requires two objects to run, the Sampler object and the Data object, both of which can be obtained
    from the Acid object. If one or the other is not provided, some methods will not work.
    """

    def __init__(
            self,
            data                    : Data|object,
            sampler                 : EnsembleSampler|None  = None,
            process_results         : bool                  = True,
            verbose                 : IntLike|bool|str|None = None,
        ) -> None:
        """
        Initialize the Result class

        Parameters
        ----------
        data : :py:class:`Data` | :py:class:`Acid`
            An Acid object or Data object (contained in Acid class). If an Acid 
            object is provided, all other arguments are taken from there. If a Data object is 
            provided, a sampler can be provided in the second argument. If a sampler object 
            is provided, it will be used as the sampler, but all other attributes will need 
            to be set manually for the Result object to be fully functional.
        sampler : :py:class:`emcee.EnsembleSampler`, optional
            Sets and overwrites the sampler in the Data object with this if provided, by default None. 
        process_results : bool, optional
            Whether to process the results from the Acid object upon initialisation, by default True.
            If False, the profiles attribute will not be available until Result.process_results() is called.
            The process_results functions does a LSD call, which can be skipped to save time and use
            the Result object for methods that do not require the profiles attribute, such as 
            continue_sampling() or plot_walkers(). This requires a Data object with the necessary attributes, 
            and a sampler object in the initialisation, or an Acid object with the necessary attributes already set.
            By default, None.
        verbose : :py:type:`IntLike | bool | str`, optional
            Verbosity level, works exactly the same as :py:class:`Acid`, if not provided
            defaults to provided :py:class:`Acid`/:py:class:`Data` class verbosity (which itself defaults to 2).
            Overwrites any value passed trough the Data object.
        """
        # Handle the different possible cases for 1st argument input
        from .acid import Acid
        if isinstance(data, Acid):
            self.data = data.data
        elif isinstance(data, Data):
            self.data = data
        else:
            raise ValueError(f"First argument must be either an Acid or Data object. Got {type(data)} instead.")

        # Handle config and verbose options
        self.config = self.data.config # point Result.config to Data.config to keep them in sync
        self.config.verbose = verbose # property overwrites or handles if verbose input was None

        # By default set sampler_initialized = False until sampler has been initialised in function so that self.initiate_sampler can be skipped
        self.sampler_initialized = False

        # Handle the sampler if input, initiate if one exists
        self.sampler = sampler if sampler is not None else self.sampler # update sampler if provided, otherwise keep the same
        if self.sampler is not None:
            self.initiate_sampler(self.sampler) # set internal variables based on sampler, sets sampler_initialized to True

        if not self.data.complete:
            if process_results:
                if self.sampler is None:
                    raise ValueError("Cannot process results without a sampler. Please provide a sampler in the initialisation or set process_results=False.")
                else:
                    self.process_results()
            elif self.config.verbose > 0:
                print("Warning: Results not processed. Profiles attribute will not be available until " \
                "Result.process_results() is called or passed through a method.")
        elif self.sampler is None and self.config.verbose>0:
            print(f"Warning: No sampler provided or found in Data object. \n" \
            f"Some methods will not work unless a sampler is provided as a parameter or if Result.initiate_sampler(sampler) is called.")

    @_require_sampler
    def process_results(self) -> None:
        """
        Processes the MCMC sampler results to obtain the final LSD profiles and continuum fit, and errors on both.
        This is effectively the final step in the ACID pipeline, and must be run before the profiles attribute is available. 
        This is automatically called if process_results is True during :py:class:`Acid` initialization.
        This function is stored here instead of the Acid class because it is not necessary to have the final profiles to use some of the
        methods contained within this class.
        """
        t0 = time()

        # Obtain flattened samples
        flat_samples = self.sampler.get_chain(discard=self.burnin, thin=self.thin, flat=True)

        # Getting the final profile and continuum values
        nvel = len(self.data.velocities) if self.config.deterministic_profile is False else 0
        quartiles = np.percentile(flat_samples, [16, 50, 84], axis=0)
        errors = np.diff(quartiles, axis=0)
        errors = np.max(errors, axis=0)
        poly_cos      = quartiles[1, nvel:]
        poly_cos_err  = errors[nvel:] # unused for now

        if self.config.verbose > 1:
            print('Getting the final profiles...')

        # Finding error for the continuum fit
        norm_wl = self.data.wavelengths["combined_normalized"]
        coeffs = flat_samples[:, nvel:]
        ncoeffs = coeffs.shape[1]
        powers = np.vander(norm_wl, N=ncoeffs, increasing=True)

        # First check memory to see if all samples can be used
        available_memory = utils.get_available_memory() # in bytes
        m_available = available_memory * 0.8 / (1024**3) # in GB, with 0.8 factor safety gap
        n_samples, ncoeffs = coeffs.shape
        npix = powers.shape[0]
        matrix_size_gb = (2 * n_samples * npix + n_samples * ncoeffs + npix * ncoeffs) * 8 / (1024**3)
        # If memory exceeded, fallback to using 1000 random samples
        if matrix_size_gb > m_available:
            if self.config.verbose > 1:
                print(f"Warning: Calculating continuum error with all samples may exceed available memory ({matrix_size_gb:.2f} GB required, {m_available:.2f} GB available). "
                "Calculating with a max of 1000 random samples instead.")
            indices_size = min(1000, n_samples)
            random_indices = np.random.choice(n_samples, size=indices_size, replace=False)
            coeffs = coeffs[random_indices, :]

        conts = (coeffs @ powers.T)
        continuum_error = np.std(conts, axis=0)

        # First get the combined profile, and then calculate each frame's profile if there are multiple frames.
        # If there is one frame, then the combined_profile is the same as the single frame profile.
        nframes = len(self.data.flux["input"])
        profiles = [] # switch to list format to add covariance matrix to result
        for counter in range(nframes+1):
            if counter == 0:
                flux = np.copy(self.data.flux["combined"])
                error = np.copy(self.data.errors["combined"])
                wavelengths = np.copy(self.data.wavelengths["combined"])
                sn = np.copy(self.data.sn["combined"])
                error[self.data.residual_masks] = 1e12
            else:
                flux = np.copy(self.data.flux["input"][counter-1])[self.data.nanmask]
                error = np.copy(self.data.errors["input"][counter-1])[self.data.nanmask]
                wavelengths = np.copy(self.data.wavelengths["input"][counter-1])[self.data.nanmask]
                sn = np.copy(self.data.sn["input"][counter-1])

                # Masking based off residuals interpolated onto new wavelength grid
                reference_wave = self.data.wavelengths["input"][np.nanargmax(self.data.sn["input"])]
                reference_wave = reference_wave[self.data.nanmask]

                reference_mask = np.zeros_like(reference_wave, dtype=bool)
                reference_mask[self.data.residual_masks] = True
                reference_interp1d = interp1d(reference_wave, reference_mask.astype(float), kind="nearest", bounds_error=False, fill_value=0.0)
                interpolated_mask = reference_interp1d(wavelengths) > 0.5
                error[interpolated_mask] = 1e12

            # Build continuum model
            a, b = utils.get_normalisation_coeffs(wavelengths)
            norm_wavelengths = (a*wavelengths)+b
            mdl = P.polyval(norm_wavelengths, poly_cos)

            # correcting continuum
            error = np.sqrt((error/mdl)**2 + (continuum_error/mdl)**2)
            flux /= mdl

            # Check whether we can skip alpha by reusing the same alpha, only true if the wavelength grid is identical
            condition = np.array_equal(wavelengths, self.data.wavelengths["combined"])
            alpha = self.data.alpha if condition else None

            LSD_profiles = LSD(self.data)
            LSD_profiles.run_LSD(wavelengths, flux, error, sn=sn, alpha=alpha)

            profile_f = LSD_profiles.profile_F
            profile_errors_f = LSD_profiles.profile_errors_F
            cov_z_f = LSD_profiles.cov_z_F

            if counter == 0:
                self.data.combined_profile = [profile_f, profile_errors_f, cov_z_f]
                self.data.continuum_model = mdl
            else:
                profiles.append([profile_f, profile_errors_f, cov_z_f])

        self.data.profiles = profiles # point Data.profiles to Result.profiles to keep them in sync
        self.data.results_time = time() - t0
        self.data.total_time = self.data.setup_time + self.data.mcmc_time + self.data.results_time
        self.data.complete = True

        # Now that results are complete, save the data instance if specified
        if self.config.save_path is not None:
            self.save() # the sampler is already saved if specified

        return

    @_require_profiles
    def __getitem__(self, item) -> list|np.ndarray:
        """
        Allows indexing into the profiles array directly from the Result object.
        """
        if isinstance(item, tuple):
            # Tuples allow for array-like indexing of the list
            if len(item) == 3:
                _order, frame, velocity = item
                return self.data.alphaprofiles[frame][velocity]
            elif len(item) == 2:
                return self.data.profiles[item[0]][item[1]]
            elif len(item) == 1:
                return self.data.combined_profile[item[0]]
            else:
                raise ValueError(f"Tuple indexing must be of length 1, 2, or 3. Got {len(item)} instead.")
        elif isinstance(item, int):
            # Return just the profile or error (or cov_mat) for single int input
            if item < 0 or item > 2:
                raise ValueError(f"Integer index must be 0, 1, or 2 to specify whether to return the profile, error, or covariance matrix. Got {item} instead.")
            return self.data.combined_profile[item]
        elif isinstance(item, str):
            # Various different options for string inputs, why not
            if "error" in item.lower():
                return self.data.combined_profile[1]
            elif "cov" in item.lower():
                return self.data.combined_profile[2]
            elif "profile" in item.lower():
                return self.data.combined_profile[0]
            else:
                raise ValueError(f"String index must contain either 'error', 'cov', or 'profile' to specify which to return. Got {item} instead.")
        else:
            raise ValueError(f"Invalid index type. Must be either a tuple, int, or str. Got {type(item)} instead.")

    @_require_profiles
    def __iter__(self):
        """Allows iterating over the profiles array directly from the Result object."""
        return iter(self.data.profiles)

    def __repr__(self):
        # Only print out the sampler and data attributes, and whether profiles is available, to avoid printing large arrays
        return f"Result object with sampler={self.sampler}, data={self.data}, profiles={'available' if self.data.profiles is not None else 'not available'}"

    def __str__(self):
        return self.__repr__()

    @_require_sampler
    def continue_sampling(self, process_results:bool=True, sampler:EnsembleSampler|None=None, **kwargs) -> None:
        """
        Continue MCMC sampling for additional steps. Passes the stored sampler into a Acid instance with the saved data. See
        :py:function:`Acid.continue_sampling` for more details on the parameters that can be passed.

        Parameters
        ----------
        process_results : bool, optional
            Whether to process the results after continuing sampling, by default True.
            If False, the profiles attribute will not be updated until Result.process_results() is called.
        sampler : emcee.EnsembleSampler | None, optional
            Optionally provide a different sampler to continue sampling from, otherwise,
            takes the sampler from the Result object, by default None
        nsteps : :py:type:`IntLike`, optional
            Number of additional MCMC steps to run. Passed to :py:function:`Acid.continue_sampling` through **kwargs.
        max_steps : :py:type:`IntLike`, optional
            Maximum number of MCMC steps to run, by default None. Passed to :py:function:`Acid.continue_sampling` through **kwargs.
        max_steps_kwargs : dict, optional
            Additional keyword arguments to be passed to the run_mcmc_until_converged function if max_steps is specified, by default None.
            The kwargs description can be found in Acid.ACID(), they are the 4 kwargs appearing after max_steps. Typos for kwargs are silently
            ignored. Passed to :py:function:`Acid.continue_sampling` through **kwargs.
        parallel : bool, optional
            Overwrites config with whether to run the MCMC in parallel. If None, uses already existing configuration. Default is None.
            Passed to :py:function:`Acid.continue_sampling` through **kwargs.
        cores : int, optional
            Overwrites config with the number of cores to use for parallel MCMC. If None, uses already existing configuration. Default is None.
            Passed to :py:function:`Acid.continue_sampling` through **kwargs.
        moves : dict, optional
            Overwrites config with the dictionary specifying the moves to use for MCMC sampling. If None, uses already existing configuration. 
            Default is None. See :py:function:`Acid.ACID` for format. Passed to :py:function:`Acid.continue_sampling` through **kwargs.
        """
        # Note that sampler input updates the sampler stored using the @_require_sampler decorator

        # Continue sampling using the Acid method
        from .acid import Acid
        acid = Acid(data=self.data) # includes config data
        acid.continue_sampling(**kwargs) # updates self.data.sampler (or self.sampler)

        self.initiate_sampler(self.sampler) # update internal variables to match new sampler

        if process_results:
            self.process_results() # update profiles
        else:
            if self.config.verbose>0:
                print("Warning: Results not processed. profiles attribute will not be available until " \
                "Result.process_results() is called.")

    @_require_sampler
    def plot_walkers(
        self,
        sampler    : EnsembleSampler|None = None,
        burnin     : IntLike|None         = None,
        thin       : IntLike|None         = None,
        return_fig : bool                 = False
        ) -> None | tuple:
        """Plots, at maximum, the last 10 MCMC walkers for the LSD profile and continuum 
        polynomial coefficients.

        Parameters
        ----------
        sampler : :py:class:`emcee.EnsembleSampler`, optional
            Optionally provide a different sampler to plot from, otherwise,
            takes the sampler from the Result object, by default None
        burnin : :py:type:`IntLike`, optional
            Optionally define the number of burnin steps, by default uses the burnin calculated when the sampler was initiated.
        thin : :py:type:`IntLike`, optional
            Optionally define the number of thinning steps, by default uses the thinning calculated when the sampler was initiated.
        return_fig : bool, optional
            Whether to return the figure and axis objects instead of showing the plot, by default False

        Returns
        ----------
        If return_fig is True, returns a tuple (fig, ax) of the figure and axes objects containing, else None
        """
        # Set burnin and thin to defaults if not provided
        burnin = burnin if burnin is not None else self.burnin
        thin = thin if thin is not None else self.thin
        samples = self.sampler.get_chain(thin=int(thin))
        steps = np.arange(samples.shape[0]) * thin

        # Setup plot and plot the walkers for the default parameters
        naxes = len(self.default_params)
        fig, ax = plt.subplots(naxes, 1, figsize=(10, 20), sharex=True)
        for i in range(naxes):
            ax[i].plot(steps, samples[:, :, self.default_params[i]], "k", alpha=0.3)
            ax[i].axvspan(0, burnin, color="red", alpha=0.1, label="burn-in")
            ax[i].set_ylabel(self.default_param_labels[i])
        ax[-1].legend()
        ax[-1].set_xlabel("Step number")
        ax[-1].set_xlim(0, self.data.nsteps)
        ax[0].set_title('MCMC Walkers')
        plt.subplots_adjust(hspace=0.05)
        if return_fig:
            return fig, ax
        plt.show()

    @_require_sampler
    def plot_corner(
        self,
        sampler    :EnsembleSampler|None = None,
        return_fig :bool                 = False,
        **kwargs,
        ) -> None | tuple:
        """Creates a corner plot for at maximum the last 8 LSD profile and continuum polynomial coefficients.

        Parameters
        ----------
        sampler : emcee.EnsembleSampler | None, optional
            Optionally provide a different sampler to plot from, otherwise,
            takes the sampler from the Result object, by default None
        return_fig : bool, optional
            Whether to return the figure object instead of showing the plot, by default False
        **kwargs:
            Additional keyword arguments to pass to corner.corner().
        
        Returns
        ----------
        If return_fig is True, returns the figure object containing the corner plot, else None
        """

        # Get samples and thin and burnin from the class variables
        samples = self.sampler.get_chain()
        samples = self.sampler.get_chain(discard=self.burnin, flat=True, thin=self.thin)[:, self.default_params]

        # Use corner.corner to handle corner plot
        fig = corner.corner(samples, labels=self.default_param_labels, show_title=True, title_fmt=".3f", title_kwargs={"fontsize": 16}, **kwargs)
        plt.suptitle('MCMC Corner Plot')
        if return_fig:
            return fig
        plt.show()

    @_require_profiles
    def plot_profiles(
        self,
        grid            :bool      = True,
        labels          :dict|None = None,
        return_fig      :bool      = False,
        subplot_kwargs  :dict|None = None,
        errorbar_kwargs :dict|None = None,
        fig_ax                     = None,
        ) -> None | tuple:
        """Plots the LSD profile result from Acid.

        Parameters
        ----------
        grid : bool, optional
            Show or hide grid, by default True
        labels : dict | None, optional
            Keys: 'xlabel', 'ylabel', and 'title'. Allows label overrides., by default None
        return_fig : bool, optional
            Whether to return the figure and axis objects instead of showing the plot, by default False
        subplot_kwargs : dict | None, optional
            Keyword arguments to be passed to plt.subplots(), by default None
        errorbar_kwargs : dict | None, optional
            Keyword arguments to be passed to ax.errorbar(), by default None
        fig_ax : tuple[matplotlib.figure.Figure, matplotlib.axes.Axes] | None, optional
            Optionally provide an existing fig/axis tuple to plot on, by default None
        
        Returns
        ----------
        If return_fig is True, returns a tuple (fig, ax) of the figure and axes objects containing the plot.
        Otherwise, displays the plot and returns None.
        """
        # Set default errorbar kwargs
        errorbar_defaults = {
            "fmt"      : ".-",
            "ecolor"   : "red",
            "linewidth": 1,
        }
        errorbar_kwargs = utils.set_dict_defaults(errorbar_kwargs, errorbar_defaults)

        # Set default subplot kwargs
        subplot_kwargs = utils.set_dict_defaults(subplot_kwargs, {"figsize": (10, 6)})

        # Set default labels
        default_labels = {
            "title" : "Acid Profile",
            "xlabel": "Velocity (km/s)",
            "ylabel": "Normalised Flux"
        }
        labels = utils.set_dict_defaults(labels, default_labels)

        # Set useful variables
        nframes = len(self.data.profiles)
        if fig_ax is None:
            fig, ax = plt.subplots(**subplot_kwargs)
        else:
            fig, ax = fig_ax

        # Iterate through and plot frames
        for f, frame in enumerate(self.data.profiles):
            x, y, yerr = self.data.velocities, frame[0], frame[1]
            label_default = f"Frame {f+1}" if nframes > 1 else None
            # Override label in errorbar_kwargs if it is not already set, otherwise use the default label
            if "label" not in errorbar_kwargs:
                errorbar_kwargs["label"] = label_default
            ax.errorbar(x, y-1, yerr=yerr, **errorbar_kwargs)

        # Add labels and titles
        ax.set_title(labels["title"])
        ax.set_xlabel(labels["xlabel"])
        ax.set_ylabel(labels["ylabel"])
        ax.axhline(0, color='black', linestyle='--', linewidth=1)
        ax.legend()
        ax.grid(grid)
        if return_fig:
            return fig, ax
        else:
            plt.show()

    @_require_profiles
    def plot_forward_model(
        self,
        grid            :bool      = True,
        labels          :dict|None = None,
        return_fig      :bool      = False,
        subplot_kwargs  :dict|None = None,
        ) -> None | tuple:
        """
        Plots the forward model calculated from the final profiles to the combined input spectrum.

        Parameters
        ----------
        grid : bool, optional
            Show or hide grid, by default True
        labels : dict | None, optional
            Keys: 'xlabel', 'ylabel', 'title', and 'residuals_ylabel'. Allows label overrides, by default None
        return_fig : bool, optional
            Whether to return the figure and axis objects instead of showing the plot, by default False
        subplot_kwargs : dict | None, optional
            Keyword arguments to be passed to plt.subplots(). Allows label overrides, by default None
        
        Returns
        ----------
        If return_fig is True, returns a tuple (fig, ax) of the figure and axes objects containing the plot.
        Otherwise, displays the plot and returns None.
        """
        
        # Set default labels
        default_labels = {
            "title"           : "Forward Model Fit to Observed Spectrum",
            "xlabel"          : "Wavelength (Angstroms)",
            "ylabel"          : "Normalised Flux",
            "residuals_ylabel": "Residuals",
        }
        labels = utils.set_dict_defaults(labels, default_labels)

        # Set default subplot kwargs
        subplot_kwargs = {
            "figsize": (10, 8),
            "sharex": True,
            "gridspec_kw": {'height_ratios': [3, 1]}
        }
        subplot_kwargs = utils.set_dict_defaults(subplot_kwargs, {"figsize": (10, 8)})

        # Get input data
        wavelengths = self.data.wavelengths["combined"]
        flux = self.data.flux["combined"]

        # Get flat_samples which are the same samples used to calculate the final profile, alpha is OD, 
        # so convert profile back to OD and reconvert to flux for forward model
        profile = utils.flux_to_od(self.data.combined_profile[0])
        model_flux = utils.od_to_flux(self.data.alpha @ profile) * self.data.continuum_model

        # Plotting
        fig, ax = plt.subplots(2, 1, **subplot_kwargs)
        ax[0].plot(wavelengths, flux, color='black', linewidth=1, label='Observed Spectrum')
        ax[0].plot(wavelengths, model_flux, color='C0', linewidth=1, label='Forward Model Fit')
        ax[0].plot(wavelengths, self.data.continuum_model, color='C1', linewidth=1, label='Fitted Continuum', linestyle='--')
        ax[1].plot(wavelengths, model_flux - flux, color='C0', linewidth=1, label='Residuals')
        ax[1].axhline(0, color='black', linestyle='--', linewidth=1)
        ax[0].set_title(labels["title"])
        ax[1].set_xlabel(labels["xlabel"])
        ax[0].set_ylabel(labels["ylabel"])
        ax[1].set_ylabel(labels["residuals_ylabel"])
        ax[1].axhline(0, color='black', linestyle='--', linewidth=1)
        ax[0].legend()
        ax[1].legend()
        ax[0].grid(grid)
        ax[1].grid(grid)
        plt.subplots_adjust(hspace=0.05)

        if return_fig:
            return fig, ax
        else:
            plt.show()

    @_require_sampler
    def plot_autocorrelation(
        self,
        sampler        : EnsembleSampler|None = None,
        n_grid         : IntLike              = 12,
        c              : float                = 5.0,
        return_fig     : bool                 = False,
        subplot_kwargs : dict|None            = None,
        min_steps      : IntLike              = 100
        ) -> None | tuple:
        """
        Plot estimated integrated autocorrelation time as a function of chain length.

        From the emcee docs:
        - For several prefixes of the chain, estimate tau with Sokal windowing.
        - Plot tau(N) and the reference line tau = N/50.

        Parameters
        ----------
        sampler : :py:class:`emcee.EnsembleSampler` | None, optional
            Optionally provide a different sampler to plot from, otherwise,
            takes the sampler from the Result object, by default None
        n_grid : :py:type:`IntLike`, optional
            Number of N values (prefix lengths) to evaluate, by default 12.
        c : float, optional
            Sokal window constant, by default 5.0.
        return_fig : bool, optional
            Whether to return the figure and axes objects, by default False
        subplot_kwargs : dict | None, optional
            Keyword arguments to be passed to plt.subplots(). Allows label overrides, by default None
        min_steps : :py:type:`IntLike`, optional
            Minimum number of post-burnin samples required to attempt autocorrelation estimation, by default 100
            If you decrease this, you may get unreliable estimates or errors from the autocorrelation time estimation.

        Returns
        ----------
        If return_fig is True, returns a tuple (fig, ax) of the figure and axes objects containing 
        the plot. Otherwise, displays the plot and returns None.
        """

        chain = self.sampler.get_chain()  # (nsteps, nwalkers, ndim)
        nsteps, nwalkers, ndim = chain.shape

        if nsteps < min_steps:
            raise ValueError("Not enough post-burnin samples to estimate autocorrelation reliably.")
        
        Ns = np.unique(np.exp(np.linspace(np.log(min_steps), np.log(nsteps), n_grid)).astype(int))
        Ns = Ns[Ns >= min_steps]  # Ensure we only consider N >= min_steps

        tau_estimates = {p: np.full(len(Ns), np.nan, dtype=float) for p in self.default_params}

        # Estimate taus
        for i, n in enumerate(Ns):
            for p in self.default_params:
                y = chain[:n, :, p].T
                tau_estimates[p][i] = utils.autocorr_new(y, c=c)

        subplot_kwargs = {} if subplot_kwargs is None else dict(subplot_kwargs)
        subplot_kwargs = utils.set_dict_defaults(subplot_kwargs, {"figsize": (10, 6)})

        fig, ax = plt.subplots(**subplot_kwargs)

        for label, p in zip(self.default_param_labels, self.default_params):
            ax.loglog(Ns, tau_estimates[p], "o-", label=f"{label}")

        # Reference line tau = N/50
        ax.loglog(Ns, Ns / 50.0, "--", label=r"$\tau = N/50$")

        ax.set_xlabel("Number of post-burnin samples per walker (N)")
        ax.set_ylabel(r"Estimated integrated autocorrelation time $\tau$")
        ax.set_title("Autocorrelation time estimates vs chain length")
        ax.legend()
        ax.grid(True, which="both")

        if return_fig:
            return fig, ax
        plt.show()

        return

    @_require_sampler
    def plot_acf(
        self,
        sampler        : EnsembleSampler|None = None,
        max_lag        : IntLike|None         = None,
        return_fig     : bool                 = False,
        subplot_kwargs : dict|None            = None,
        ) -> None | tuple:
        """
        Plot the autocorrelation function (ACF) for each parameter, averaged across walkers.
        
        Parameters
        ----------
        sampler : :py:class:`emcee.EnsembleSampler`, optional
            Optionally provide a different sampler to plot from, otherwise,
            takes the sampler from the Result object, by default None
        max_lag : :py:type:`IntLike`, optional
            Maximum lag to plot, by default None (plots up to min(5000, nsteps-1))
        return_fig : bool, optional
            Whether to return the figure and axes objects, by default False
        subplot_kwargs : dict, optional
            Keyword arguments to be passed to plt.subplots(). Allows label overrides, by default None

        Returns
        -------
        If return_fig is True, returns a tuple (fig, ax) of the figure and axes objects containing 
        the plot. Otherwise, displays the plot and returns None.
        """
        chain = self.sampler.get_chain() 
        nsteps, nwalkers, ndim = chain.shape

        subplot_kwargs = {} if subplot_kwargs is None else dict(subplot_kwargs)
        subplot_kwargs = utils.set_dict_defaults(subplot_kwargs, {"figsize": (10, 5)})

        fig, ax = plt.subplots(**subplot_kwargs)

        for param, label in zip(self.default_params, self.default_param_labels):

            y = chain[:, :, param].T  # (nwalkers, nsteps)

            # Mean ACF across walkers
            f = np.zeros(nsteps)
            for w in range(nwalkers):
                f += utils.autocorr_func_1d(y[w], norm=True)
            f /= nwalkers

            if max_lag is None:
                max_lag = min(5_000, nsteps - 1)
            max_lag = int(max_lag)

            ax.plot(np.arange(max_lag + 1), f[: max_lag + 1], label=f"{label}")
        
        ax.set_xlabel("Lag (steps)")
        ax.set_ylabel("Autocorrelation")
        ax.set_title(f"Mean ACF across walkers")
        ax.set_xscale("log")
        ax.grid(True)
        ax.axhline(0, color="black", linestyle="--", linewidth=1)
        ax.legend()

        if return_fig:
            return fig, ax
        plt.show()

    def initiate_sampler(self, sampler:EnsembleSampler|None, _method_name=None) -> None:
        """
        Initiates the sampler attribute from an external sampler.

        Parameters
        ----------
        sampler : :py:class:`emcee.EnsembleSampler`
            An emcee EnsembleSampler object to set as the sampler attribute.
        _method_name : str, optional
            Internal parameter used to track which method is calling initiate_sampler, for error messages. 
            Not intended for user input, by default None.
        """
        if self.sampler_initialized:
            if sampler is None:
                return # sampler already initiated from initialisation, so skip the rest of the method
            # else: continues to update the sampler and internal variables based on new sampler input
        self.sampler = sampler if sampler is not None else self.sampler
        if self.sampler is None:
            if _method_name is not None:
                error_msg = f"Cannot run {_method_name} without a sampler, please pass in a sampler to the method or during initialisation."
            else:
                error_msg = "Cannot initiate sampler without a sampler stored in the instance or passed as a parameter, please pass in a sampler "
            raise AttributeError(error_msg)

        # Calculate autocorr time, burnin, thin
        # Suppress output from get_autocorr_time call
        with open(os.devnull, "w") as devnull, \
            contextlib.redirect_stdout(devnull), \
            contextlib.redirect_stderr(devnull):
            self.tau = self.sampler.get_autocorr_time(quiet=True)

        self.converged = True
        if self.data.nsteps < 50 * np.max(self.tau):
            self.converged = False
            if self.config.verbose>1:
                print("The number of MCMC steps is less than 50 times the maximum autocorrelation " \
                "time.\n The sampler may not have converged. Consider running more steps or checking " \
                f"the walker plots.\n The max autocorrelation time is {np.max(self.tau):.2f}, therefore " \
                f"the minimum number of steps should be roughly {int(50 * np.max(self.tau))}.\n Disabling burnin " \
                f"from autocorrelation time, instead using burnin=steps-1000")

        try:
            self.thin = int(np.min(self.tau)/5)
            if self.converged:
                self.burnin = int(3 * np.max(self.tau))
            else:
                self.burnin = self.data.nsteps - 1000 # just the last 1000 steps
        except:
            if self.config.verbose>0:
                print(f"Warning: Could not compute autocorrelation time for burnin and thinning.\n This is likely" \
                f" due to all posterior samples being rejected (possibly by prior constraints).\n The resulting profile is likely" \
                f" wrong. Setting defaults: burnin=nsteps-1000, and thin=1.")
            self.burnin = self.data.nsteps - 1000 # just the last 1000 steps
            self.thin = 1
        
        self.burnin = int(np.clip(self.burnin, 0, self.data.nsteps-1)) # ensure burnin is at least 0 and less than total steps
        self.thin = int(np.clip(self.thin, 1, self.data.nsteps-1)) # ensure thin is at least 1, and not clipping to nsteps

        # Below is used for the parameters for the walker and corner plots
        n_poly_params = self.config.poly_ord + 1
        poly_params = np.arange(-1, -n_poly_params-1, -1).tolist()
        
        # Generates labels for the polynomial coefficients, starting from 'a' for the highest order term, and going backwards in the alphabet.
        a=ord('a')
        alph=[chr(i) for i in range(a,a+26)]
        poly_labels = [alph[i] for i in range(n_poly_params)]
        
        samples = self.sampler.get_chain(thin=self.thin, discard=self.burnin)
        if not self.config.deterministic_profile:
            max_profile_idx = np.argmax(samples[:,:,:-n_poly_params].mean(axis=(0,1)))
            poly_params.extend([-5, max_profile_idx, 1])
            poly_labels.extend(["$Z_{-1}$", "$Z_{max}$", "$Z_0$"])

        self.default_params = poly_params
        self.default_param_labels = poly_labels

        self.sampler_initialized = True

    @property
    def sampler(self) -> EnsembleSampler|None:
        """Returns the sampler attribute, by default is None if not saved."""
        return self.data.sampler

    @sampler.setter
    def sampler(self, value: EnsembleSampler|None) -> None:
        """Sets the sampler in the data class."""
        self.data.sampler = value

    def save(self, *args, **kwargs) -> None:
        """
        Saves the Data instance which the Result class inherits from Acid.
        See :py:function:`Data.save` for more details on the parameters that can be passed.
        """
        self.data.save(*args, **kwargs)

    @classmethod
    def load(cls, data:str|Result|Data="result.pkl") -> Result:
        """
        Loads a Result object from a pickle file or from a Data/Result object.

        Parameters
        ----------
        data : str | :py:class:`Result` | :py:class:`Data`, optional
            A pickle file name or an object with the same attributes as a saved Result object, by default "result.pkl"

        Returns
        ----------
        :py:class:`Result`
            A Result object loaded from the pickle file or from the provided object.
        """
        if isinstance(data, str):
            return cls(Data.load(data))
        elif isinstance(data, Result):
            return cls(data.data)
        elif isinstance(data, Data):
            return cls(data)



from __future__ import annotations
import numpy as np
from astropy.io import  fits
import glob, psutil, os
import matplotlib.pyplot as plt
from scipy.signal import find_peaks
from scipy.interpolate import LSQUnivariateSpline
from tqdm import tqdm
from scipy.linalg import cho_factor, cho_solve
from beartype import beartype
from . import utils
from .errors import LineListRangeError, SNCutError
from .data import Config, Data, LineList
from .utils import c_kms, IntLike, Scalar, Array1D, Array2D

@beartype
class LSD:
    """
    Class containing all useful functions for performing least-squares deconvolution.
    This does not simultaneously fit the continuum and perform LSD (which ACID does). It is used
    for the initial parameters of the ACID mcmc run and for obtaining final profiles. It 
    can also be used as a standalone LSD implementation and for trying to do LSD without OD.
    For more details and an example, see :ref:`LSD` in the documentation.
    """
    def __init__(
            self,
            data    : object|None           = None,
            OD      : bool                  = True,
            verbose : IntLike|bool|str|None = None,
            ) -> None:
        """Initialises the LSD class, optionally with a Data instance to take parameters from.

        Parameters
        ----------
        data : object | None, optional
            A data instance to draw parameters and configs from, by default None
        OD : bool, optional
            Whether to perform LSD in optical depth space (True) or flux space (False), by default True.
            We generally recommend always using optical depth as ACID was always intended, but you can set
            this to False if you wish to do your own testing. See :ref:`LSD` in the documentation for more details.
        verbose : :py:type:`IntLike | bool | str | None`, optional
            Verbosity level, if None, uses the :py:class:`Config` class existing value (in Data), or default of 2.
            Should follow the same format as :py:class:`Acid` verbosity. 
            Will overwrite the verbosity level in the config if a Data instance is input, by default None.
        """
        # Set class variables, taking from input data if it exists, else setting to defaults
        self.slurm    = "SLURM_JOB_ID" in os.environ
        self.data     = data if data is not None else Data()
        self.linelist = data.linelist if data is not None else None
        self.OD       = OD
        try:
            self.config = data.config
        except:
            self.config = Config() # uses defaults
        self.config.update_hipri(verbose=verbose) # Update config with new values, if not None

    def run_LSD(
        self,
        wavelengths : Array1D,
        flux        : Array1D,
        errors      : Array1D,
        sn          : Scalar,
        linelist    : Array2D|str|LineList|dict|None = None,
        velocities  : Array1D|None                   = None,
        alpha       : Array2D|None                   = None,
        ) -> None:
        """Runs the LSD algorithm to extract the average line profile from the observed spectrum.

        Parameters
        ----------
        wavelengths : :py:type:`Array1D`
            Array of wavelengths of the observed spectrum in Angstroms
        flux : :py:type:`Array1D`
            Array of flux values corresponding to the wavelengths (in linear space, and should be continuum normalized)
        errors : :py:type:`Array1D`
            Array of error values corresponding to the flux
        sn : :py:type:`Scalar`
            Signal-to-noise ratio of the observed spectrum
        linelist : :py:type:`Array2D | str | LineList | dict | None`, optional
            Linelist to use for LSD, should follow the same format as :py:class:`Acid`. 
            If None, uses the linelist already stored in the class, if it exists, by default None.
        velocities : :py:type:`Array1D`, optional
            Array of velocities corresponding to the observed spectrum.
            If the class was not initialised with an Acid instance, this is required, by default None
        alpha : :py:type:`Array2D` | None, optional
            Precomputed alpha matrix, if already calculated and you want to skip directly to the Cholesky 
            decomposition and solving for the profile, by default None
        """
        # Ensure inputs are numpy arrays
        wavelengths = np.asarray(wavelengths)
        flux = np.asarray(flux)
        errors = np.asarray(errors)

        # Ensure dimensions match
        if not wavelengths.shape == flux.shape == errors.shape:
            raise ValueError("Input wavelengths, flux, and errors must have the same shape.")       

        # Set linelist and velocities either from inputs or from Data class if initialised with Acid instance
        self.data.velocities = velocities if velocities is not None else self.data.velocities
        if self.data.velocities is None:
            raise ValueError("Velocities must be provided either as an argument to run_LSD or when initialising the class with an Acid instance.")
        self.data.linelist = linelist # Raises if no linelist available, overwrites if input

        # Unpack the linelist stored in data
        wavelengths_linelist, depths_linelist = self.data.linelist

        # Clip linelist to wavelength range of spectrum
        wavelengths_linelist, depths_linelist = utils.clip_wavelengths(wavelengths, wavelengths_linelist, depths_linelist)
        if len(wavelengths_linelist) == 0:
            raise LineListRangeError(f"No lines in linelist are within the wavelength range of the observed spectrum. \n"\
                                     f"You may have mismatched wavelengths units between linelist and spectrum or an empty linelist.\n"\
                                     f"Please check your linelist and input spectrum.")

        # Apply S/N cut (of 1/(3*SN)) to linelist
        wavelengths_linelist, depths_linelist = self.sn_clip(wavelengths_linelist, depths_linelist, sn)

        # Convert to optical depth space for the linelist and the spectrum if needed, and convert errors accordingly
        if self.OD:
            flux, errors, depths_linelist = utils.flux_to_od(flux, errors, depths_linelist)

        # Calculates alpha in optical depth, selects lines greater than 1/(3*sn)
        if alpha is None:
            self.alpha = self.calc_alpha(wavelengths, wavelengths_linelist, depths_linelist)
        else:
            self.alpha = alpha

        # Now solve for profile using Cholesky decomposition
        self.c_factor = self.calc_cholesky(self.alpha, errors)

        # Solve for profile and profile errors using Cholesky factors
        self.profile, self.profile_errors, self.cov_z = self.solve_z(self.alpha, flux, errors, self.c_factor, return_cov=True)

        # Convert profile back to flux if needed
        if self.OD:
            self.profile_F, self.profile_errors_F, self.cov_z_F = utils.od_to_flux(self.profile, self.profile_errors, cov_matrix=self.cov_z)
        else:
            self.profile_F, self.profile_errors_F, self.cov_z_F = self.profile, self.profile_errors, self.cov_z

        return

    def sn_clip(
            self,
            wavelengths_linelist : Array1D,
            depths_linelist      : Array1D,
            sn                   : Scalar,
            ) -> tuple[Array1D, Array1D]:
        """
        Applies a signal-to-noise cut to the linelist, removing lines shallower than 1/(3*sn) as per Dolan et al (2024).

        Parameters
        ----------
        wavelengths_linelist : :py:type:`Array1D`
            Wavelengths from the linelist
        depths_linelist : :py:type:`Array1D`
            Depths from the linelist
        sn : :py:type:`Scalar`
            Signal-to-noise ratio threshold

        Returns
        -------
        tuple[:py:type:`Array1D`, :py:type:`Array1D`]
            Clipped wavelengths and depths from the linelist
        """
        # Selecting lines deeper than 1/(3*sn)
        idx = (depths_linelist >= 1/(3*sn))
        wavelengths_linelist = wavelengths_linelist[idx]
        depths_linelist = depths_linelist[idx]

        # Analyse remaining lines
        ncut = np.sum(~idx)
        nrest = np.sum(idx)
        perc = 100 * nrest / (nrest + ncut)
        if nrest == 0:
            raise SNCutError(f"No lines remain in the linelist after S/N cut. Please check your linelist and S/N value.")
        if self.config.verbose > 0:
            if perc < 5:
                print("Warning: Less than 5% of lines remain after S/N cut. Check your linelist and S/N value.")
            if self.config.verbose > 2:
                print(f"{perc:.2f}% of lines used in LSD: {nrest} out of {nrest + ncut} remain from S/N cut.")
        return wavelengths_linelist, depths_linelist

    def calc_alpha(
        self,
        wavelengths          : Array1D,
        wavelengths_linelist : Array1D,
        depths_linelist      : Array1D,
        velocities           : Array1D|None = None,
        ) -> Array2D:
        """
        Calculates the alpha matrix given flux and errors and a linelist.
        Note that if this function is called without using run_LSD, there is no selection of lines deeper than 1/(3*sn).
        If you still wish to do this, it needs to be done in linear space with the sn_clip function before converting to (if desired) OD space.
        The units of the alpha matrix will match the units of the input linelist.

        Parameters
        ----------
        wavelengths : :py:type:`Array1D`
            Array of wavelengths of the observed spectrum in optical depth space
        wavelengths_linelist : :py:type:`Array1D`
            Array of wavelengths from the linelist in optical depth space
        depths_linelist : :py:type:`Array1D`
            Array of depths from the linelist in optical depth space
        velocities : :py:type:`Array1D`, optional
            Array of velocities, needs to either be initialised by class with Acid instance, or input here, by default None
        
        Returns
        -------
        :py:type:`Array2D`
            The alpha matrix, to be used in the Cholesky decomposition and solving for the profile. 
            The units will match the units of the input linelist (OD or flux).
        """

        # Set velocities from input or from class variable if initialised with Acid instance, raise error if not provided in either way
        if velocities is not None:
            self.data.velocities = velocities
        elif self.data.velocities is None:
            raise ValueError("Velocities must be provided either as input or initialized in the Acid instance.")

        # Check velocity spacing is constant
        if not np.allclose(np.diff(self.data.velocities), self.data.velocities[1] - self.data.velocities[0]):
            raise ValueError("Velocity spacing must be constant for the alpha matrix calculation.")

        # Calculate velocity pixel size
        deltav = self.data.velocities[1] - self.data.velocities[0]

        # Clip linelist to wavelength range of spectrum (again just in case this is called without run_LSD, saves memory by reducing lines)
        wavelengths_linelist, depths_linelist = utils.clip_wavelengths(wavelengths, wavelengths_linelist, depths_linelist)

        # Find differences and velocities
        blankwaves = wavelengths
        diff = blankwaves[:, np.newaxis] - wavelengths_linelist
        vel = c_kms * (diff / wavelengths_linelist)

        # Get memory available depnding on whether were on slurm or not
        available_memory = utils.get_available_memory() # in bytes
        mat_size = len(wavelengths_linelist) * len(self.data.velocities) * len(blankwaves) * 8 # Matrix size in bytes
        m_available = available_memory / 2  # Available memory in bytes (divided by 2 to be safe)

        # Calculate alpha matrix in one go if it fits in memory
        if mat_size < m_available:
            # Calculating entire alpha matrix at once, broadcasts into a 3D array of shape (n_wl, n_lines, n_vel)
            x = (vel[:, :, np.newaxis] - self.data.velocities) / deltav
            delta = np.clip(1.0 - np.abs(x), 0.0, 1.0)
            alpha = (depths_linelist[:, None] * delta).sum(axis=1)  # (n_wl, n_vel)

        # Else, calculate in blocks to save memory
        else:
            n_blank = len(blankwaves)
            n_vel   = len(self.data.velocities)
            mem_size = available_memory // 2
            bytes_per_row = n_blank * n_vel * 8 * 3 # *8 for float64, *3 for vel, x, delta in a row
            max_block = max(1, mem_size // bytes_per_row)
            block = int(min(max_block, len(wavelengths_linelist)))
            # Set initial alpha matrix to np.zeros
            alpha  = np.zeros((len(blankwaves), len(self.data.velocities)), dtype=np.float64)

            # Use tqdm progress bar if verbose
            if self.config.verbose>1:
                iterable = tqdm(range(0, len(wavelengths_linelist), block), desc='Calculating alpha matrix')
            else:
                iterable = range(0, len(wavelengths_linelist), block)

            for start_pos in iterable:
                # Ensure we don't go out of bounds on last iteration
                end_pos = min(start_pos + block, len(wavelengths_linelist))
                wl  = wavelengths_linelist[start_pos:end_pos]
                dep = depths_linelist[start_pos:end_pos]

                # Perform calculations for this block
                vel_blk = c_kms * (blankwaves[:, None] - wl) / wl
                x_blk   = (vel_blk[:, :, None] - self.data.velocities) / deltav
                delta   = np.clip(1.0 - np.abs(x_blk), 0.0, 1.0)                    

                alpha += (dep[:, None] * delta).sum(axis=1)
        return alpha

    @staticmethod
    def calc_cholesky(
        alpha : Array2D,
        error : Array1D,
        **kwargs,
        ) -> tuple:
        """
        Calculates the LHS Cholesky factorisation matrix given the alpha matrix and flux errors.
        The units of alpha and error should match (ie OD or flux), the output will be in the same units.

        Parameters
        ----------
        alpha : :py:type:`Array2D`
            The precomputed alpha matrix
        error : :py:type:`Array1D`
            Flux errors
        **kwargs : dict, optional
            Additional keyword arguments to pass to scipy.linalg.cho_factor. 
            Overwrite_a=False must be set by us, do not pass this as a kwarg.

        Returns
        -------
        c_factor : tuple
            Cholesky factorisation matrix and lower/upper flag, to be put straight into solve_z as `c_factor` parameter
        """
        V = 1.0 / (error ** 2) # variance vector in log space, error already in log space

        # M = αT V α,  b = αT V R
        AVA = alpha.T @ (V[:, None] * alpha)

        # Diangostics for common 1-th leading order linalg error
        # M = alpha.T @ (V[:, None] * alpha)
        # print("finite M:", np.all(np.isfinite(M)))
        # print("min diag:", np.min(np.diag(M)))
        # print("rank:", np.linalg.matrix_rank(M), " / ", M.shape[0])
        # col_norm = np.linalg.norm(np.sqrt(V)[:, None] * alpha, axis=0)
        # print("zero columns:", np.sum(col_norm == 0))

        # Cholesky factorisation of M
        c_factor = cho_factor(AVA, overwrite_a=False, **kwargs)
        return c_factor

    @staticmethod
    def solve_z(
        alpha        : Array2D,
        flux         : Array1D,
        error        : Array1D,
        c_factor     : tuple,
        return_error : bool = True,
        return_cov   : bool = False,
        **kwargs,
        ) -> tuple:
        """
        Solves for the LSD profile and its errors using the Cholesky factors. 
        All units should match between alpha, flux, and error (ie all in OD or all in flux).
        Returns the profile in the same units.

        Parameters
        ----------
        alpha : :py:type:`Array2D`
            The precomputed alpha matrix
        flux : :py:type:`Array1D`
            The observed flux values in optical depth space
        error : :py:type:`Array1D`
            The flux errors in optical depth space
        c_factor : tuple
            Cholesky factorisation matrix and lower/upper flag, to be put straight into 
            scipy.linalg.cho_solve as `c_factor` parameter
        return_error : bool, optional
            Whether to calculate and return the profile errors along with the
            profile, by default True
        return_cov : bool, optional
            Whether to return the full covariance matrix instead of just the errors, by default False
        **kwargs : dict, optional
            Additional keyword arguments to pass to both scipy.linalg.cho_solve calls
            (one for the profile, one for the covariance matrix)
            Overwrite_b=False is already set by default, do not pass this as a kwarg.

        Returns
        -------
        profile, profile_errors, cov_z : tuple
            LSD profile and its errors (if return_error is True) and covariance matrix (if return_cov is True)
        """
        V = 1.0 / (error ** 2) # variance vector in log space, error already in log space
        R = flux         # R matrix in log space

        # M = αT V α,  b = αT V R
        AVR = alpha.T @ (V * R)

        # z = M⁻¹ b
        profile = cho_solve(c_factor, AVR, overwrite_b=False, **kwargs)

        # Find error, cov(z) = M⁻¹, take diagonal
        if return_error or return_cov:
            AVA = alpha.T @ (V[:, None] * alpha)
            cov_z = cho_solve(c_factor, np.eye(AVA.shape[0]), overwrite_b=False, **kwargs)
            profile_errors = np.sqrt(np.diag(cov_z))
            if return_cov:
                return profile, profile_errors, cov_z
            else:
                return profile, profile_errors
        else:
            return profile

    @classmethod
    def convolve_profile(
        cls,
        profile              : Array1D,
        profile_errors       : Array1D,
        alpha                : Array2D|None = None,
        velocities           : Array1D|None = None,
        wavelengths          : Array1D|None = None,
        linelist_wavelengths : Array1D|None = None,
        linelist_depths      : Array1D|None = None,
        ) -> tuple[Array1D, Array1D]:
        """
        Convolve your profile either using an inputted alpha matrix or by calculating one using :py:meth:`calc_alpha` 
        with the inputted wavelengths and linelist. The units of the output convolved model spectrum will match the 
        units of the input profile (ie OD or flux) and alpha matrix/linelist depths. If alpha is not input, the wavelengths 
        and linelist inputs are required to calculate the alpha matrix.
        See :py:func:`utils.flux_to_od` and :py:func:`utils.od_to_flux` for conversions.

        Parameters
        ----------
        profile : :py:type:`Array1D`
            The LSD profile to be convolved, in the same units as the alpha matrix (OD or flux)
        profile_errors : :py:type:`Array1D`
            The errors on the LSD profile, in the same units as the alpha matrix (OD or flux)
        alpha : :py:type:`Array2D` | None, optional
            Precomputed alpha matrix, if already calculated and you want to skip directly to the convolution, by default None
        velocities : :py:type:`Array1D` | None, optional
            Array of velocities corresponding to the observed spectrum, required if alpha is not input, by default None
        wavelengths : :py:type:`Array1D` | None, optional
            Array of wavelengths of the observed spectrum, required if alpha is not input, by default None
        linelist_wavelengths : :py:type:`Array1D` | None, optional
            Array of wavelengths from the linelist, required if alpha is not input, by default None
        linelist_depths : :py:type:`Array1D` | None, optional
            Array of depths from the linelist, required if alpha is not input. Must be in the same units
            as the alpha matrix (OD or flux), by default None
        """

        if alpha is None:
            if velocities is None or wavelengths is None or linelist_wavelengths is None or linelist_depths is None:
                raise ValueError("If alpha matrix is not input, velocities, wavelengths, linelist_wavelengths, and " \
                "linelist_depths must all be provided to calculate the alpha matrix.")
            cls.__init__(cls)
            alpha = cls.calc_alpha(cls, wavelengths, linelist_wavelengths, linelist_depths, velocities, verbose=2)

        model_spectrum = alpha @ profile
        model_errors = np.sqrt((alpha**2) @ (profile_errors**2))

        return model_spectrum, model_errors
